import { getDateTime } from './utils'

/** @enum LoggerLevels 日志输出等级 */
export enum LoggerLevels {
  error,
  warn,
  log,
  debug
}

declare global {
  interface Window {
    ctiLoggerLevel?: LoggerLevels
    HS_CTI_CLIENT_ID?: string
  }
}
/** @class Logger 通用的日志模块 */
export class Logger {
  private _level: LoggerLevels
  private category: string
  private label: string | undefined
  constructor(level: LoggerLevels, category: string, label?: string) {
    this._level = level
    this.category = category
    this.label = label
  }

  public error(content: string): void {
    this.genericLog(LoggerLevels.error, content)
  }
  public warn(content: string): void {
    this.genericLog(LoggerLevels.warn, content)
  }
  public log(content: string): void {
    this.genericLog(LoggerLevels.log, content)
  }
  public debug(content: string): void {
    this.genericLog(LoggerLevels.debug, content)
  }

  public genericLog(levelToLog: LoggerLevels, content: string): void {
    if (this._level >= levelToLog) {
      this.print(levelToLog, this.category, this.label, content)
    }
  }

  private print(
    levelToLog: LoggerLevels,
    category: string,
    label: string | undefined,
    content: string
  ): void {
    if (typeof content === 'string') {
      const prefix: Array<string> = [getDateTime()]
      if (label) {
        prefix.push(label)
      }
      content = prefix.concat(content).join(' | ')
    }
    switch (levelToLog) {
      case LoggerLevels.error:
        console.error(`%c${category}`, 'color: blue;', content)
        break
      case LoggerLevels.warn:
        console.warn(`%c${category}`, 'color: blue;', content)
        break
      case LoggerLevels.log:
        console.log(`%c${category}`, 'color: blue;', content)
        break
      case LoggerLevels.debug:
        console.debug(`%c${category}`, 'color: blue;', content)
        break
      default:
        break
    }
  }
  get level(): LoggerLevels {
    return this._level
  }
  set level(newLevel: LoggerLevels) {
    if (newLevel >= 0 && newLevel <= 3) {
      this._level = newLevel
    } else if (newLevel > 3) {
      this._level = 3
      // eslint-disable-next-line no-prototype-builtins
    } else if (LoggerLevels.hasOwnProperty(newLevel)) {
      this._level = newLevel
    } else {
      this.error("invalid 'level' parameter value: " + JSON.stringify(newLevel))
    }
  }
}
