import { SimpleUserPlus } from './SImpleUserPlus'
import EventEmitter from '../eventemitter'
import { Logger, LoggerLevels } from '../Logger'
import {
  Scene,
  type HsCTIInitOptions,
  type RequiredHsCTIInitOptions,
  CTIStatus,
  SIPStatus,
  SocketStatus,
  CTIEvent,
  CTIErrorType,
  HskTerminatedCode,
  SdkErrorCode,
  CTIRes,
  CallStatus
} from './outputType'
import type {
  InitOptions,
  CTIBaseOptions,
  CTIManualCallOptions
} from '../api/hs-cti/ctiSdkModel'
import HsSocket from './HsSocket'
import {
  setBaseOption,
  resetBaseOption,
  BaseOption,
  getClientId
} from './storage'
import {
  AudioName,
  ExceptMessage,
  type SIPStatusChangeParams,
  type SocketStatusChangeParams,
  SocketEvent,
  CTIEventParams,
  SimpleUserPlusDelegate,
  SimpleUserPlusOptions
} from './type'

import { upperCamelToLowerSnake } from '@/utils'
import {
  agentCheckIn,
  // agentCheckOut,
  agentSetBusy,
  agentSetIdle,
  getAgentStatus,
  getCtiFlowId,
  getInitConf,
  listen,
  loadAgentGroupData,
  manualCall,
  manualHang,
  setActiveServiceTask
} from '../api/hs-cti/ctiSdk'
import {
  checkCTIStatus,
  getUserMedia,
  handleApiRes,
  validateParams
} from './tools'
import { Invitation, SessionState, UserAgentState } from 'sip.js'

/**
 * 本地提示音
 * _ringAudio 机器人外呼/监听，等待接起提示音
 * _waitAudio 主动外呼，点击拨打时的等待音
 * _byeAudio 结束通话提示音
 */
const audioList = {
  _ringAudio: 'https://static.fuxicarbon.com/hs-cti/ring.wav',
  _waitAudio: 'https://static.fuxicarbon.com/hs-cti/manual.wav',
  _byeAudio: 'https://static.fuxicarbon.com/hs-cti/bye.wav'
}

/** @class HsCTI 红杉外呼类 */

export class HsCTI extends EventEmitter {
  private logger: Logger
  public loggerLevel: LoggerLevels
  /** HsCTI 实例 */
  public static instance: HsCTI | undefined
  public scene: Scene
  public agentId: string
  public vccId: string

  /** 接口返回的初始化配置 */
  private _initOptions: InitOptions | undefined

  /** IM socket的实例 */
  private _socket!: HsSocket
  /** sip.js UA实例 */
  private _sipUserAgent: SimpleUserPlus | undefined

  private _callId!: string
  private _ctiFlowIdList!: string[]

  /** 基本参数 */
  private _baseParams: CTIBaseOptions
  /** 等待提示音播放器 */
  private _waitAudio: HTMLAudioElement
  /** 振铃提示音播放器 */
  private _ringAudio: HTMLAudioElement
  /** 结束通话提示音 */
  private _byeAudio: HTMLAudioElement
  /** 远端音频流播放器 */
  private _remoteAudio: HTMLAudioElement
  /** CTI状态 */
  private _ctiStatus!: CTIStatus
  private _ctiStatusList!: CTIStatus[]
  /** sip状态 */
  private _sipStatus!: SIPStatus
  private _sipStatusList!: SIPStatus[]
  /** socket状态 */
  private _socketStatus!: SocketStatus
  private _socketStatusList!: SocketStatus[]

  private _callStatus: CallStatus
  private _terminatedStatusList: string[]

  private constructor(
    hsCTIInitOptions: HsCTIInitOptions | RequiredHsCTIInitOptions
  ) {
    const { vccId, agentId, scene, password, env, loggerLevel } =
      hsCTIInitOptions
    super()
    this.loggerLevel = window.ctiLoggerLevel || loggerLevel || LoggerLevels.log
    this.logger = new Logger(this.loggerLevel, 'HsCTI')
    this._waitAudio = new Audio()
    this._ringAudio = new Audio()
    this._byeAudio = new Audio()
    this._remoteAudio = new Audio()

    this.vccId = vccId
    this.agentId = agentId
    this.scene = scene
    this._callId = ''
    setBaseOption(BaseOption.ENV, env)
    setBaseOption(BaseOption.LoggerLevel, this.loggerLevel)

    this._baseParams = { agentId, vccId, scene, password }
    const baseTrackParams = { agent_id: agentId, vcc_id: vccId, scene }
    if (scene === Scene.Monitor) {
      const { monitorScene } = hsCTIInitOptions
      this._baseParams = { ...this._baseParams, monitorScene }
      setBaseOption(
        BaseOption.TrackParams,
        { ...baseTrackParams, monitor_scene: monitorScene },
        true
      )
    } else {
      setBaseOption(BaseOption.TrackParams, baseTrackParams, true)
    }

    this.initInstanceOptions()
    this.setCTIStatus(CTIStatus.Initial)
    // 首次获取实例时，将外呼状态置为未开始
    this._callStatus = CallStatus.Stopped
    this._terminatedStatusList = []
  }

  get getCTIStatus() {
    return this._ctiStatus
  }
  get getSocketStatus() {
    return this._socketStatus
  }
  get getSIPStatus() {
    return this._sipStatus
  }

  @validateParams()
  public static getInstance(
    hsCTIInitOptions: HsCTIInitOptions | RequiredHsCTIInitOptions
  ): HsCTI {
    if (!HsCTI.instance) {
      HsCTI.instance = new HsCTI(hsCTIInitOptions)
    }
    setBaseOption(BaseOption.TrackParams, {
      clientId: getClientId()
    })
    return HsCTI.instance
  }

  /** @private 重置实例 */
  private initInstanceOptions() {
    this._ctiFlowIdList = []
    this._ctiStatusList = []
    this._sipStatusList = []
    this._socketStatusList = []
    HsCTI.instance = undefined
    this._initOptions = undefined
    resetBaseOption()
  }

  /** @private 优雅关闭 SIP 和 socket */
  private clearSocketAndSip() {
    this._socket?.closeSocket()
    this._sipUserAgent?.unregister()
    this._sipUserAgent?.disconnect()
    this._sipUserAgent = undefined
  }

  /**
   * @private 设置等待音 src
   * @param {AudioName} audioName
   * @param {boolean} loop
   */
  private setAudioSrc(audioName: AudioName, loop: boolean) {
    this.logger.debug(`media | 设置等待音 src: ${audioList[audioName]}`)
    this[audioName].src = audioList[audioName]
    this[audioName].currentTime = 0
    this[audioName].autoplay = false
    this[audioName].loop = loop
  }

  private playAudio(audioName: AudioName) {
    this.logger.debug(`media | 播放音频：${audioName}`)
    this[audioName].play()
  }

  private stopLocalAudio() {
    switch (this.scene) {
      case Scene.Robot:
      case Scene.Monitor:
        this.stopAudio(AudioName.RingAudio, true)
        break
      case Scene.Manual:
        this.stopAudio(AudioName.WaitAudio, true)
        break
    }
  }

  private stopAudio(audioName: AudioName, loop: boolean) {
    this.logger.debug(`media | 停止音频播放：${audioName}`)
    this[audioName].src = ''
    setTimeout(() => {
      this.setAudioSrc(audioName, loop)
    }, 1000)
  }

  @getUserMedia()
  public init() {
    // 不允许重复初始化，必须在销毁sdk后才能重新初始化
    const [lastStatus] = this._ctiStatusList.slice(-1)
    if (lastStatus !== CTIStatus.Initial) return
    this.setAudioSrc(AudioName.ByeAudio, false)
    switch (this.scene) {
      case Scene.Robot:
      case Scene.Monitor:
        this.setAudioSrc(AudioName.RingAudio, true)
        break
      case Scene.Manual:
        this.setAudioSrc(AudioName.WaitAudio, true)
        break
    }
    this.getInitConfig()
  }

  private async getInitConfig() {
    const res = await getInitConf(this._baseParams)
    const { code, data: initOptions, msg } = res
    /**
     * 由于后端接口层面对于查询来说，查不到不代表异常，所以即使查不到也是正向流程
     * 因此如果传入一个不存在的agentId，后端也会返回code=0，但是data=null
     * 这里需要根据code===0&&initOptions存在进行联合判断，只有两个条件都成立，才表示该坐席获取fs以及im等配置成功
     */
    if (code === 0 && initOptions) {
      setBaseOption(BaseOption.TrackParams, {
        sip_server: initOptions.sipServer,
        cti_session_id: initOptions.ctiSessionId
      })
      this._initOptions = initOptions
      this.initSocket(initOptions)
      this.initSip(initOptions)
    } else {
      this.eventEmitAndTrack(CTIEvent.OnCtiError, {
        type: CTIErrorType.ServerTerminated,
        code: HskTerminatedCode.GetInitConfig,
        // 如果code===0并且initOptions不存在时，代表未查到对应坐席，此时接口返回的msg是空字符串，所以提示文案要特定处理
        msg: code === 0 && !initOptions ? '未查到坐席' : msg,
        method: 'getInitConfig',
        res_code: code
      })
    }
  }

  private initSocket(initOptions: InitOptions) {
    this.setSocketStatus({ status: SocketStatus.Initial })
    this._socket = new HsSocket({
      agentId: this.agentId,
      vccId: this.vccId,
      // appCode: initOptions.appCode,
      imWsServer: initOptions.imWsServer,
      imWsPath: initOptions.imWsPath,
      imHeartTime: initOptions.imHeartTime,
      imRetryCount: initOptions.imRetryCount,
      ctiSessionId: initOptions.ctiSessionId,
      loggerLevel: this.loggerLevel
    })
    this._socket.on(SocketEvent.SocketDownEvent, ({ eventData }) => {
      const { eventName, ext } = eventData
      this.logger.log(
        `socket server down | ${eventName} | ${JSON.stringify(ext)}`
      )

      this.handleSocketDownEvent({ eventName, ext })
    })

    this._socket.on(SocketEvent.SetSocketStatus, res => {
      this.setSocketStatus(res)
    })

    this._socket.initSocket()
  }

  private handleSocketDownEvent({ eventName, ext }: CTIEventParams) {
    const ctiFlowId = this._baseParams.ctiFlowId
    const extCtiFlowId = ext.ctiFlowId
    if (
      [Scene.Manual, Scene.Monitor].includes(this.scene) &&
      ctiFlowId &&
      extCtiFlowId &&
      ctiFlowId !== extCtiFlowId
    ) {
      this.logger.error(
        `cti_flow_id | 不一致! fe: ${ctiFlowId}, server: ${extCtiFlowId}, eventName: ${eventName}`
      )
      return
    }

    this.serverEventEmit({ eventName, ext })
  }

  /** @private serverEventEmit 统一处理服务端推送的事件 */
  private serverEventEmit({ eventName, ext }: CTIEventParams) {
    const NO_EMIT_EVENT_LIST = [
      CTIEvent.OnRingStart,
      CTIEvent.OnDetectedTone,
      CTIEvent.OnRingEnd
    ]
    switch (eventName) {
      case CTIEvent.OnRingStart:
        this.stopLocalAudio()
        break
      case CTIEvent.OnDetectedTone:
        this.stopLocalAudio()
        break
      case CTIEvent.OnAgentWorkReport:
        break
      case CTIEvent.OnCallEnd:
        /** 通话结束：重置可拨打状态，清空本轮通话的CallId */
        this._callStatus = CallStatus.Stopped
        setBaseOption(BaseOption.TrackParams, {
          call_id: ''
        })
        break
      case CTIEvent.OnRingEnd:
        /**  TODO: 后 4 个事件未来服务端不再推送时删掉  */
        break
      default:
      // do nothing
    }

    if (!NO_EMIT_EVENT_LIST.includes(eventName)) {
      this.eventEmitAndTrack(eventName, ext)
    }
  }

  private initSip(initOptions: InitOptions) {
    this.setSipStatus({ status: SIPStatus.Initial })
    const server = initOptions.wssServer
    const simpleUserPlusOptions: SimpleUserPlusOptions = {
      aor: initOptions.sipServer,
      reconnectionAttempts: initOptions.fsRetryCount,
      reconnectionDelay: initOptions.fsRetryTime,
      optionsPingInterval: initOptions.fsHeartTime,
      media: {
        remote: {
          audio: this._remoteAudio
        }
      },
      userAgentOptions: {
        gracefulShutdown: true,
        authorizationPassword: initOptions.phonePwd,
        sessionDescriptionHandlerFactoryOptions: {
          constraints: {
            audio: {
              echoCancellation: true,
              noiseSuppression: true,
              autoGainControl: true
            },
            video: false
          },
          peerConnectionConfiguration: {
            iceServers: []
          }
        }
      }
    }

    simpleUserPlusOptions.delegate = this.sipDelegate()

    this._sipUserAgent = new SimpleUserPlus(server, simpleUserPlusOptions)
    this._sipUserAgent.connect().catch((error: Error) => {
      const err = `SIP UserAgent 启动失败：${JSON.stringify(error)}`
      this.logger.error(err)
      this.setSipStatus({
        status: SIPStatus.Terminated,
        code: HskTerminatedCode.SIPInitUserAgent,
        error: err,
        method: 'initSIPJS'
      })
    })
  }

  private sipDelegate(): SimpleUserPlusDelegate {
    return {
      onServerConnect: () => {
        this._sipUserAgent!.register()
      },
      onServerDisconnect: () => {
        // 正向关闭UA时会进入此回调
        this.setSipStatus({ status: SIPStatus.Terminated })
      },
      onReconnectStart: () => {
        // 开始重连时将状态流转为ReTry
        this.setSipStatus({ status: SIPStatus.ReTry })
      },
      onReconnectFailed: () => {
        // 重连失败时会进如此回调
        this.setSipStatus({
          status: SIPStatus.Terminated,
          code: HskTerminatedCode.SIPOnDisconnect,
          error: `超过重连次数，终止重连`,
          method: 'onReconnectFailed'
        })
        this.setSipStatus({ status: SIPStatus.Terminated })
      },
      onRegistered: () => {
        this.setSipStatus({ status: SIPStatus.Ready })
      },
      onUserAgentStateChange: (state: UserAgentState) => {
        switch (state) {
          case UserAgentState.Started:
            this.setSipStatus({ status: SIPStatus.Started })
            break
          case UserAgentState.Stopped:
            this.setSipStatus({
              status: SIPStatus.Terminated,
              code: this._sipStatusList.includes(SIPStatus.Started)
                ? HskTerminatedCode.SIPUserAgentStateStopped
                : HskTerminatedCode.SIPInitUserAgent,
              error: 'SIP UserAgentState 状态停止',
              method: 'onUserAgentStateChange'
            })
            break
        }
      },
      onInvite: (invitation: Invitation) => {
        const callId = invitation.request.getHeader('P-Wdh-Callid') || ''
        const ctiFlowId = invitation.request.getHeader('P-WDH-CtiFlowId') || ''
        if (
          this.scene === Scene.Manual &&
          ctiFlowId !== this._baseParams.ctiFlowId
        ) {
          this.logger.error(
            `cti_flow_id 不一致! fe: ${this._baseParams.ctiFlowId} | P-WDH-CtiFlowId: ${ctiFlowId}`
          )
          return
        }
        this._callId = callId
        setBaseOption(BaseOption.TrackParams, {
          call_id: callId
        })
        this.sessionStateChangeAndTrack(invitation.state)
        if ([Scene.Robot, Scene.Monitor].includes(this.scene)) {
          this.playAudio(AudioName.RingAudio)
        }

        if ([Scene.Manual].includes(this.scene)) {
          this.answer()
        }

        invitation.delegate = {
          onCancel: cancel => {
            const error = `sip_invitation_on_cancel | ${JSON.stringify(cancel)}`
            this.logger.error(error)
            this.eventEmitAndTrack(
              CTIEvent.OnCtiError,
              {
                type: CTIErrorType.SdkError,
                code: SdkErrorCode.InvitationCancel,
                msg: '当前通话已结束',
                method: 'sipDelegate'
              },
              error
            )
          }
        }

        invitation.stateChange.addListener((state: SessionState) => {
          this.sessionStateChangeAndTrack(state)

          switch (state) {
            case SessionState.Initial:
            case SessionState.Establishing:
              break
            case SessionState.Established:
              if ([Scene.Robot, Scene.Monitor].includes(this.scene)) {
                this.stopLocalAudio()
              }
              break
            case SessionState.Terminating:
              break
            case SessionState.Terminated:
              this.playAudio(AudioName.ByeAudio)
              setTimeout(() => {
                this.stopAudio(AudioName.ByeAudio, false)
              }, 1000)
              this.stopLocalAudio()
              this._callStatus = CallStatus.Stopped
              setBaseOption(BaseOption.TrackParams, {
                call_id: ''
              })
              break
            default:
              break
          }
        })
      }
    }
  }

  private sessionStateChangeAndTrack(status: SessionState) {
    this.eventEmitAndTrack(CTIEvent.OnSessionStatusChange, {
      status
    })
    const trackName = `sip_session_state_${upperCamelToLowerSnake(status)}`
    this.logger.log(trackName)
  }

  /**
   * @private setCTIStatus CTI 状态流转
   * @param {CTIStatus} ctiStatus
   */
  private setCTIStatus(ctiStatus: CTIStatus) {
    if (ctiStatus === this.getCTIStatus) return

    this._ctiStatus = ctiStatus
    this._ctiStatusList.push(ctiStatus)
    const logInfo = `cti status | ${ctiStatus} | ${this._ctiStatusList}`
    if (ctiStatus === CTIStatus.Terminated) {
      this.setSocketStatus({ status: SocketStatus.Terminated })
      this.setSipStatus({ status: SIPStatus.Terminated })
      this.stopLocalAudio()
      this.logger.warn(logInfo)
    } else {
      this.logger.debug(logInfo)
    }
    if (ctiStatus === CTIStatus.Ready) {
      this.setSocketStatus({ status: SocketStatus.Ready })
      this.setSipStatus({ status: SIPStatus.Ready })
      this._terminatedStatusList = []
      if (!this._ctiStatusList.includes(CTIStatus.ReTry)) {
        this.checkIn()
      }
    }
  }

  /**
   * @private setSocketStatus Socket 状态流转
   * @param SocketStatusChangeParams Socket 状态流转参数及错误详情等
   */
  private setSocketStatus({ status, code, error }: SocketStatusChangeParams) {
    if (status === this.getSocketStatus) return

    this._socketStatus = status
    this._socketStatusList.push(status)
    const logInfo = `socket status | ${status} | ${this._socketStatusList}`
    status === SocketStatus.Terminated
      ? this.logger.warn(logInfo)
      : this.logger.debug(logInfo)

    this.socketOrSipStatusChange(status, this.getSIPStatus)
    if (error) {
      this._terminatedStatusList.push(code as string)
      this.eventEmitAndTrack(
        CTIEvent.OnCtiError,
        {
          type: CTIErrorType.SdkTerminated,
          code,
          msg:
            code === HskTerminatedCode.SocketRepeatLogin
              ? ExceptMessage.CTIRepeatLoginMsg
              : ExceptMessage.CommonNetworkErrorMsg,
          method: 'setSocketStatus'
        },
        {
          error_msg: error,
          socket_status_list: this._socketStatusList,
          sip_status_list: this._sipStatusList,
          cti_status_list: this._ctiStatusList
        }
      )
      this.clearSocketAndSip()
    }
  }

  /**
   * @private setSipStatus SIP 状态流转
   * @param SIPStatusChangeParams SIP 状态流转参数及错误详情等
   */
  private setSipStatus({ status, code, error, method }: SIPStatusChangeParams) {
    if (status === this.getSIPStatus) return

    this._sipStatus = status
    this._sipStatusList.push(status)
    const logInfo = `sip status | ${status} | ${this._sipStatusList}`
    status === SIPStatus.Terminated
      ? this.logger.warn(logInfo)
      : this.logger.debug(logInfo)

    this.socketOrSipStatusChange(this.getSocketStatus, status)

    if (error) {
      /** 抛出 SIP 类型的错误详情 */
      this._terminatedStatusList.push(code as string)
      this.eventEmitAndTrack(
        CTIEvent.OnCtiError,
        {
          type: CTIErrorType.SdkTerminated,
          code,
          msg: ExceptMessage.CommonNetworkErrorMsg,
          method: method || 'setSipStatus'
        },
        {
          error_msg: error,
          socket_status_list: this._socketStatusList,
          sip_status_list: this._sipStatusList,
          cti_status_list: this._ctiStatusList
        }
      )
      this.clearSocketAndSip()
    }
  }

  /**
   * @private socketOrSipStatusChange Socket 或 SIP 状态变化可能引起 CTI 状态变化
   * @param {SocketStatus} socketStatus
   * @param {SIPStatus} sipStatus
   */
  private socketOrSipStatusChange(
    socketStatus: SocketStatus,
    sipStatus: SIPStatus
  ) {
    if (socketStatus === SocketStatus.Ready && sipStatus === SIPStatus.Ready) {
      this.setCTIStatus(CTIStatus.Ready)
    }
    if (socketStatus === SocketStatus.ReTry || sipStatus === SIPStatus.ReTry) {
      this.setCTIStatus(CTIStatus.ReTry)
    }
    if (
      socketStatus === SocketStatus.Terminated ||
      sipStatus === SIPStatus.Terminated
    ) {
      this.setCTIStatus(CTIStatus.Terminated)
    }
  }

  /**
   * @private eventEmitAndTrack SDK 对外抛出事件并统一埋点
   * @param {CTIEvent} eventName 事件名称
   * @param {object} ext 事件参数
   * @param { error } error 错误详情或错误辅助信息
   */
  private eventEmitAndTrack(
    eventName: CTIEvent,
    ext: object,
    error?: string | object
  ) {
    this.logger.debug(`sdk emit | ${eventName} | ${JSON.stringify(ext)}`)
    // 如果出现异常，则将通话状态置为结束
    if (eventName === CTIEvent.OnCtiError) {
      this._callStatus = CallStatus.Stopped
    }
    try {
      this.emit(eventName, ext)
      console.log(error)
    } catch (error) {
      this.logger.error(`业务监听 ${eventName} 事件，处理回调时报错: ${error}`)
    }
  }

  /** @private checkIn 服务端签入，CTIStatus Ready 时自动调用，坐席状态变更 */
  @checkCTIStatus('签入')
  @handleApiRes()
  private async checkIn() {
    const res = await agentCheckIn(this._baseParams)
    if (res.code === 0) {
      this.eventEmitAndTrack(CTIEvent.OnInitalSuccess, {
        vccId: this.vccId,
        agentId: this.agentId,
        scene: this.scene,
        phoneNum: this._initOptions!.phoneNum,
        sipServer: this._initOptions!.sipServer
      })
    }
    return res
  }

  /** @private checkOut 服务端签出， unInit 时自动调用，坐席状态变更 */
  // @handleApiRes()
  // private async checkOut() {
  //   return await agentCheckOut(this._baseParams)
  // }

  /** @private _getCtiFlowId 获取手动外呼场景需要的 ctiFlowId */
  @handleApiRes()
  private async getCtiFlowId() {
    const res = await getCtiFlowId(this._baseParams)
    const { code, data } = res
    if (code === 0) {
      this._baseParams.ctiFlowId = data
      this._ctiFlowIdList.push(data)
      setBaseOption(BaseOption.TrackParams, {
        cti_flow_id: data,
        cti_flow_id_list: JSON.stringify(this._ctiFlowIdList)
      })
    }
    return res
  }

  /** @public setIdle 服务端置闲，坐席状态变更 */
  @checkCTIStatus('置闲')
  @handleApiRes()
  public async setIdle() {
    return await agentSetIdle(this._baseParams)
  }

  /** @public setBusy 服务端置忙，坐席状态变更 */
  @checkCTIStatus('置忙')
  @handleApiRes()
  public async setBusy() {
    return await agentSetBusy(this._baseParams)
  }

  @checkCTIStatus('人工外呼')
  public async makeCall(params: CTIManualCallOptions) {
    // 如果当前通话状态处于外呼开始，则不允许再次调用此方法并上报埋点
    if (this._callStatus === CallStatus.Started) {
      return
    }
    this._callStatus = CallStatus.Started
    await this.getCtiFlowId()
    this.playAudio(AudioName.WaitAudio)
    return await this.serverCall(params)
  }

  @handleApiRes()
  private async serverCall({ called, ext }: CTIManualCallOptions) {
    const params = { ...this._baseParams, called, ...ext }
    const res = await manualCall(params)
    const { code, data } = res
    if (code === 0) {
      if (this._callId === '' && data) {
        this._callId = data
        setBaseOption(BaseOption.TrackParams, {
          call_id: data
        })
      }
    } else {
      // 外呼接口调用失败将通话状态置为结束不阻碍下次外呼
      this._callStatus = CallStatus.Stopped
      this.stopLocalAudio()
    }
    return res
  }

  /** @public answer SDK SIP 接起 */
  @checkCTIStatus('接起')
  public answer() {
    return new Promise<CTIRes>((resolve, reject) => {
      this._sipUserAgent
        ?.answer({
          sessionDescriptionHandlerOptions: {
            constraints: {
              audio: true,
              video: false
            }
          }
        })
        .then(() => {
          resolve({
            code: 0,
            data: 'answer',
            msg: 'SIP 接起电话成功'
          })
          this.logger.debug('sip_accept_success')
        })
        .catch(err => {
          const errorData = {
            type: CTIErrorType.SdkError,
            code: SdkErrorCode.Answer,
            msg:
              this.scene === Scene.Manual
                ? ExceptMessage.ManualCallAnswerErrorMsg
                : ExceptMessage.RobotOrWeChatAnswerErrorMsg,
            method: 'answer'
          }
          reject(errorData)
          this.eventEmitAndTrack(CTIEvent.OnCtiError, errorData, `${err}`)
          this.logger.error(`${CTIEvent.OnCtiError} | ${err}`)
        })
    })
  }

  /** @public bye SDK SIP 挂断 */
  @checkCTIStatus('挂断')
  public bye() {
    return new Promise<CTIRes>((resolve, reject) => {
      this._sipUserAgent
        ?.hangup()
        .then(() => {
          resolve({
            code: 0,
            data: 'bye',
            msg: 'SIP 挂断电话成功'
          })
          this.logger.debug('sip_bye_success')
        })
        .catch(err => {
          const errorData = {
            type: CTIErrorType.SdkError,
            code: SdkErrorCode.Bye,
            msg: ExceptMessage.SipByeErrorMsg,
            method: 'bye'
          }
          reject(errorData)
          if (err.message !== 'Session does not exist.') {
            this.eventEmitAndTrack(CTIEvent.OnCtiError, errorData, `${err}`)
          }
          this.logger.error(`${CTIEvent.OnCtiError} | ${err}`)
        })
    })
  }

  /** @public serverBye 挂断且流转坐席状态 */
  public async serverBye() {
    try {
      this.bye()
      // 如果人工外呼场景下，没有flowId则不调用并上报埋点
      if (!this._baseParams.ctiFlowId && this.scene === Scene.Manual) {
        return
      }
      const res = await this.turnHang()
      return Promise.resolve(res)
    } catch (error) {
      return Promise.reject(error)
    }
  }

  /** @public turnHang 流转坐席状态-通话结束 */
  @handleApiRes()
  private async turnHang() {
    const res = await manualHang({ ...this._baseParams, callId: this._callId })
    this._callId = ''
    return res
  }

  /** @public getAgentStatus 获取坐席状态 */
  @handleApiRes()
  public async getAgentStatus() {
    return await getAgentStatus(this._baseParams)
  }

  /**
   * @public loadAgentGroupData 监听-根据监听组 ID 获取监听组成员
   * @param {string[]} monitorIds
   */
  @checkCTIStatus('获取监听组成员')
  @handleApiRes()
  public async loadAgentGroupData(monitorIds: string[]) {
    return await loadAgentGroupData({ ...this._baseParams, monitorIds })
  }

  /**
   * @public listen 监听-服务端发起监听
   * @param {string} monitoredAgNo
   */
  @checkCTIStatus('坐席监听')
  @handleApiRes()
  public async listen(monitoredAgNo: string) {
    const flowIdRes = await this.getCtiFlowId()
    if (flowIdRes && flowIdRes.code !== 0) {
      return flowIdRes
    }

    return await listen({
      ...this._baseParams,
      agentId: monitoredAgNo,
      leaderAgentId: this.agentId
    })
  }

  /**
   * @public setActiveService 机器人外呼-签入人工组
   * @param {string} serviceId
   */
  @checkCTIStatus('签入人工组')
  @handleApiRes()
  public async setActiveService(serviceId: string) {
    return await setActiveServiceTask({ ...this._baseParams, serviceId })
  }

  /** @public unInit 卸载 SDK，checkOut 成功后断开 socket 和 sip 连接，并销毁 SdCTI 实例 */
  public async unInit() {
    // await this.checkOut()
    this.setCTIStatus(CTIStatus.Terminated)
    this.initInstanceOptions()
    this.clearSocketAndSip()
    this._callId = ''
  }
}
