import { LoggerLevels } from '../Logger'
/**
 * 默认 'test'
 * 'development': 本地（不上报服埋点只打印 log）
 * 'test': 测试
 * 'production': 生产
 */
export type ENV = 'development' | 'test' | 'production'
/**
 * @enum {string} CTI 初始化场景
 * Manual: 手动外呼
 * Robot: 机器人外呼
 * Monitor: 监听
 * Predictive: 预测式外呼
 */
export enum Scene {
  Manual = 'manual',
  Robot = 'robot',
  Monitor = 'monitor',
  Predictive = 'predictive'
}

// type ExcludeSceneKey = Exclude<keyof typeof Scene, 'Monitor'>
export type ExcludeScene = Scene.Manual | Scene.Robot | Scene.Predictive

/**
 * @enum {string} CTI 监听场景
 * All: 不区分场景
 * Manual: 手动外呼
 * Robot: 机器人外呼
 * Predictive: 预测式外呼
 */
export enum MonitorScene {
  All = 'all',
  Manual = 'manual',
  Robot = 'robot',
  Predictive = 'predictive'
}

/**
 * @interface {} 初始化 SD_CTI 的基本参数
 * agentId: 坐席 ID
 * vccId: 主体标识，例如 'hongshan'
 */
export interface BaseOptions {
  agentId: string
  vccId: string
}

export interface BaseHsCTIInitOptions extends BaseOptions {
  loggerLevel?: LoggerLevels
  password: string
  env: ENV
}

export interface HsCTIInitOptions extends BaseHsCTIInitOptions {
  scene: ExcludeScene
}

export interface RequiredHsCTIInitOptions extends BaseHsCTIInitOptions {
  scene: Scene.Monitor
  monitorScene: MonitorScene
}

/**
 * @interface {} 初始化 SD_CTI 需要的参数
 * loggerLevel: 日志等级
 * password: 临时的鉴权字符串，由业务方写死传进来
 * scene: CTI 初始化场景
 * monitorScene: 监听场景
 * env: 环境变量
 */

/**
 * @enum {string} Socket 状态
 * Initial: 初始状态
 * Connecting: Socket 开始建立连接
 * Connected: Socket 建立连接成功
 * Ready: 向 IM 发送第一个上行 login 消息收到成功回调
 * Terminated: Socket 连接断开、各种 Socket 错误流转到本状态
 */
export enum SocketStatus {
  Initial = 'Initial',
  Connecting = 'Connecting',
  Connected = 'Connected',
  Ready = 'Ready',
  ReTry = 'ReTry',
  Terminated = 'Terminated'
}

/**
 * @enum {string} SIP 状态
 * Initial: 初始状态
 * Started: SIP 的 User Agent 创建成功
 * Connecting: SIP 底层 Socket 传输 TransportState.Connecting
 * Connected: SIP 底层 Socket 传输 TransportState.Connected
 * Ready: SIP Registerer 监听注册状态 RegistererState.Registered
 * Terminated: SIP Socket 断开、注册失败等各种错误、主动断开连接流转到本状态
 */
export enum SIPStatus {
  Initial = 'Initial',
  Started = 'Started',
  Connecting = 'Connecting',
  Connected = 'Connected',
  Ready = 'Ready',
  ReTry = 'ReTry',
  Terminated = 'Terminated'
}

/**
 * @enum {string} CTI 状态
 * Initial: 初始状态
 * Ready: SocketStatus Ready && SIPStatus Ready
 * Terminated: SocketStatus Terminated || SIPStatus Terminated || 正常调用 unInit 方法卸载
 */
export enum CTIStatus {
  Initial = 'Initial',
  Ready = 'Ready',
  ReTry = 'ReTry',
  Terminated = 'Terminated'
}

/**
 * @enum {string} 通话状态
 * 为了防止人工外呼方法被二次调用引发预期以外的问题，增加此状态的流转
 * Started: 外呼已开始，此状态下不允许再次发起外呼
 * Stopped: 外呼已结束，此状态下可以再次发起外呼
 */
export enum CallStatus {
  Started = 'Started',
  Stopped = 'Stopped'
}

/**
 * @enum {string} Session 状态
 * CTI 目前只有 Invitation（接受会话）的场景，Inviter (主动发起会话）暂时没有
 */
export enum SessionStatus {
  /**
   * If `Inviter`, INVITE not sent yet.
   * If `Invitation`, SDK 收到 INVITE 通话请求，但尚未处理.
   */
  Initial = 'Initial',
  /**
   * If `Inviter`, sent INVITE and waiting for a final response.
   * If `Invitation`, received INVITE and attempting to send 200 final response (but has not sent it yet).
   */
  Establishing = 'Establishing',
  /**
   * If `Inviter`, sent INVITE and received 200 final response and sent ACK.
   * If `Invitation`, SDK 完成接受 INVITE 并发送 200 OK 确认接起，同时接通本地语音流.
   */
  Established = 'Established',
  /**
   * If `Inviter`, sent INVITE, sent CANCEL and now waiting for 487 final response to ACK (or 200 to ACK & BYE).
   * If `Invitation`, received INVITE, sent 200 final response and now waiting on ACK and upon receipt will attempt BYE
   * (as the protocol specification requires, before sending a BYE we must receive the ACK - so we are waiting).
   */
  Terminating = 'Terminating',
  /**
   * If `Inviter`, sent INVITE and received non-200 final response (or sent/received BYE after receiving 200).
   * If `Invitation`, SDK 收到 BYE 信令，发送 200 OK 挂断确认，会话结束.
   */
  Terminated = 'Terminated'
}

/**
 * @enum {string} CTI 所有错误的分类
 * SdkTerminated: SDK 不可用错误，需要重新初始化
 * SdkError: SDK 状态可用，其他普通错误
 * ServerTerminated: 服务端不可用错误，需要重新初始化，透传服务端 code,msg
 * ServerError: 服务端可用，普通错误，透传服务端 code,msg
 */
export enum CTIErrorType {
  SdkTerminated = 'SdkTerminated',
  SdkError = 'SdkError',
  ServerTerminated = 'ServerTerminated',
  ServerError = 'ServerError'
}

/**
 * @enum {string} SdkTerminated 类型错误的 code 枚举
 * CTITerminated: SDK 状态不可用，CTIStatus 的状态为 Terminated
 * GetUserMedia: 获取坐席媒体权限失败
 * GetInitConfig: 调接口获取 CTI 初始化配置失败
 * SocketOnError: 监听 socket.io 的 error 事件
 * SocketOnConnectError: 监听 socket.io 的 connect_error 事件
 * SocketOnDisconnect: 监听 socket.io 的 disconnect 事件
 * SocketRepeatLogin: 多页面重复登录，IM 互踢事件
 * SIPInitUserAgent: SIP UserAgent 初始化时启动失败
 * SIPInitRegister: SIP Register 初始化时注册失败
 * SIPUserAgentStateStopped: 监听 SIP UserAgent stateChange 事件状态变更为 Stopped
 * SIPTransportStateDisconnected: 监听 SIP Transport StateChange 事件状态变更为 Disconnect
 * SIPRegistererStateTerminated: 监听 SIP Registerer StateChange 事件状态变更为 Terminated
 * SIPOnDisconnect: 监听 SIP OnDisconnect 事件收到异常退出 error
 * SIPInitTransport: SIP Transport 初始化时连接失败
 * SipHeartBeatErr: SIP发送心跳OPTIONS事件时收到异常结果
 * SIPUnRegistered: 注册SIP时失败
 * SocketOnReconnectFailed: Socket重连超过阈值且依然重连失败
 */
export enum HskTerminatedCode {
  CTITerminated = '100001',
  GetUserMedia = '100002',
  GetInitConfig = '100003',
  SocketOnError = '110001',
  SocketOnConnectError = '110002',
  SocketOnDisconnect = '110003',
  SocketRepeatLogin = '110004',
  SocketOnReconnectFailed = '110007',
  SIPInitUserAgent = '120001',
  SIPInitRegister = '120002',
  SIPUserAgentStateStopped = '120003',
  SIPTransportStateDisconnected = '120004',
  SIPRegistererStateTerminated = '120005',
  SIPOnDisconnect = '120006',
  SIPInitTransport = '120007',
  SipHeartBeatErr = '120008',
  SIPUnRegistered = '120009'
}

/**
 * @enum {string} SdkError 类型错误的 code 枚举
 * Answer: SIP accept 接起失败
 * Bye: SIP bye 挂断失败
 * InvitationCancel: SIP Invitation 会话请求被取消
 * AssignStream: 播放语音流失败
 * FetchError: 修饰器handleApiRes当进入到catch时上报此code
 */
export enum SdkErrorCode {
  Answer = '200001',
  Bye = '200002',
  InvitationCancel = '200003',
  AssignStream = '200004',
  FetchError = '200005'
}

/**
 * @enum {string} CTI 事件推送
 * OnCtiError: CTI 错误事件，含前后端所有错误，SDK 推送
 * OnSessionStatusChange: 坐席侧 SIP 会话状态变更事件，SDK 推送
 * OnInitalSuccess: CTI 初始化成功事件，SDK 推送
 * OnAgentWorkReport: 坐席&用户状态变更事件，Server 推送
 * OnRingStart: 手动外呼用户未接听时，开始播放回铃音，Server 推送
 * OnRingEnd: 手动外呼用户未接听时，播放回铃音结束，Server 推送
 * OnAgentReport: 坐席状态变更事件，Server 推送
 * OnCallReportInfo: 通话时长及通话次数等信息，Server 推送
 * OnDetectedTone: 服务端收到音频信号后推送
 */
export enum CTIEvent {
  OnCtiError = 'OnCtiError',
  OnSessionStatusChange = 'OnSessionStatusChange',
  OnInitalSuccess = 'OnInitalSuccess',
  OnAgentWorkReport = 'OnAgentWorkReport',
  OnRingStart = 'OnRingStart',
  OnRingEnd = 'OnRingEnd',
  OnDetectedTone = 'OnDetectedTone',
  OnAgentReport = 'OnAgentReport',
  OnCallReportInfo = 'OnCallReportInfo',
  //  TODO: 后 7 个事件服务端未来不再推送时删掉
  OnCallRing = 'OnCallRing',
  OnCallEnd = 'OnCallEnd',
  OnCallAnswer = 'OnCallAnswer',
  OnAgentGroupQuery = 'OnAgentGroupQuery',
  OnMethodResponseEvent = 'OnMethodResponseEvent',
  OnEventPrompt = 'OnEventPrompt',
  OnPrompt = 'OnPrompt'
}

/** @interface CTIRes CTI 对外方法暴露 promise 结果 */
export interface CTIRes {
  code: number
  // eslint-disable-next-line @typescript-eslint/no-explicit-any
  data: any
  msg: string
}
