import type { Logger } from '../Logger'
import {
  TrackSource,
  baseRequireParams,
  allRequiredParams,
  methodExceptMsgMap
} from './type'
import {
  CTIStatus,
  CTIEvent,
  CTIErrorType,
  HskTerminatedCode,
  type CTIRes,
  type HsCTIInitOptions,
  type RequiredHsCTIInitOptions,
  Scene,
  SdkErrorCode
} from './outputType'

/**
 * @function getServerErrorType 根据服务端返回的 code 生成错误类型
 * @param {number} code
 * @returns {CTIErrorType}
 */
export function getServerErrorType(code: number) {
  if (code >= 300001 && code <= 399999) return CTIErrorType.ServerTerminated
  else return CTIErrorType.ServerError
}

/** @function getUserMedia 获取媒体权限 */
export function getUserMedia() {
  // eslint-disable-next-line @typescript-eslint/no-explicit-any
  return function (_target: unknown, _methodName: string, descriptor: any) {
    const originalMethod = descriptor.value

    descriptor.value = function (...args: unknown[]) {
      if (
        navigator &&
        navigator.mediaDevices &&
        navigator.mediaDevices.getUserMedia
      ) {
        navigator.mediaDevices
          .getUserMedia({
            audio: true,
            video: false
          })
          .then(() => {
            this.logger.debug('media | getUserMedia | 获取浏览器媒体权限成功')

            return originalMethod.apply(this, args)
          })
          .catch(error => {
            this.logger.error(`media | getUserMedia | ${error}`)

            this.eventEmitAndTrack(
              CTIEvent.OnCtiError,
              {
                type: CTIErrorType.SdkTerminated,
                code: HskTerminatedCode.GetUserMedia,
                msg:
                  error.name === 'NotAllowedError'
                    ? '用户拒绝了获取麦克风权限！'
                    : '获取麦克风权限失败',
                method: 'getUserMedia'
              },
              {
                source: TrackSource.FeMedia,
                event_name: 'get_user_media_error',
                error: error.name
              }
            )
          })
      } else {
        this.logger.error(
          'media | getUserMedia | 浏览器版本过低，不支持获取媒体权限'
        )
        this.eventEmitAndTrack(
          CTIEvent.OnCtiError,
          {
            type: CTIErrorType.SdkTerminated,
            code: HskTerminatedCode.GetUserMedia,
            msg: '浏览器版本过低，不支持获取媒体权限',
            method: 'getUserMedia'
          },
          {
            source: TrackSource.FeMedia,
            event_name: 'get_user_media_not_support'
          }
        )
      }
    }

    return descriptor
  }
}

/**
 * @function assignStreamError 语音流播放错误
 * @param {string} type
 * @param {Error} error
 * @param {Logger} logger
 */
function assignStreamError(type: string, error: Error, logger: Logger) {
  const err = `Failed to ${type} media: ${error}`
  logger.error(err)
}

/**
 * @function assignStream Assign a MediaStream to an HTMLMediaElement and update if tracks change
 * @param {MediaStream} stream
 * @param {HTMLMediaElement} element
 * @param {Logger} logger
 */
export function assignStream(
  stream: MediaStream,
  element: HTMLMediaElement,
  logger: Logger
): void {
  /** Set element source. */
  element.autoplay = true
  element.srcObject = stream

  logger.log(`media | 把语音流接入到本地播放`)

  /** Load and start playback of media. */
  element.play().catch((error: Error) => {
    assignStreamError('play', error, logger)
  })

  /** If a track is added, load and restart playback of media. */
  stream.onaddtrack = (): void => {
    element.load()
    element.play().catch((error: Error) => {
      assignStreamError('onaddtrack', error, logger)
    })
  }

  /** If a track is removed, load and restart playback of media. */
  stream.onremovetrack = (): void => {
    element.load()
    element.play().catch((error: Error) => {
      assignStreamError('onremovetrack', error, logger)
    })
  }
}

/**
 * @function checkCTIStatus 校验 cti 状态
 * @param {string} msg
 */
export function checkCTIStatus(msg: string) {
  // eslint-disable-next-line @typescript-eslint/no-explicit-any
  return function (_target: unknown, methodName: string, descriptor: any) {
    const originalMethod = descriptor.value

    descriptor.value = function (...args: unknown[]) {
      /** 调用 SDK method */
      this.logger.log(`sdk method | ${methodName} | ${msg}`)

      /** 校验 CTIStatus, 如果不是 Ready 状态，直接报错 */
      if (this.getCTIStatus !== CTIStatus.Ready) {
        const errorData = {
          type: CTIErrorType.SdkTerminated,
          code: HskTerminatedCode.CTITerminated,
          msg: methodExceptMsgMap[methodName],
          method: methodName,
          terminated_source: this._terminatedStatusList
        }
        this.eventEmitAndTrack(CTIEvent.OnCtiError, errorData)
        return Promise.reject(errorData)
      }

      return originalMethod.apply(this, args)
    }

    return descriptor
  }
}
/** @function handleApiRes 统一处理服务端接口返回值 */
export function handleApiRes() {
  // eslint-disable-next-line @typescript-eslint/no-explicit-any
  return function (_target: unknown, methodName: string, descriptor: any) {
    const originalMethod = descriptor.value
    descriptor.value = async function (...args: unknown[]) {
      try {
        const res: CTIRes = await originalMethod.apply(this, args)
        const { code, msg } = res
        if (code !== 0) {
          const serverErrorType = getServerErrorType(code)
          const errorData = {
            type: serverErrorType,
            code,
            msg: msg,
            method: methodName
          }
          this.eventEmitAndTrack(CTIEvent.OnCtiError, errorData)
          if (serverErrorType === CTIErrorType.ServerTerminated) {
            this.setCTIStatus(CTIStatus.Terminated)
          }
          return Promise.reject(errorData)
        } else {
          return Promise.resolve(res)
        }
      } catch (error) {
        const errorData = {
          type: 'fetch_error',
          code: SdkErrorCode.FetchError,
          msg: JSON.stringify(error),
          method: methodName
        }
        this.eventEmitAndTrack(CTIEvent.OnCtiError, errorData)
        return Promise.reject(errorData)
      }
    }

    return descriptor
  }
}

export function validateParams() {
  // eslint-disable-next-line @typescript-eslint/no-explicit-any
  return function (_target: unknown, _methodName: string, descriptor: any) {
    const originalMethod = descriptor.value
    descriptor.value = function (
      ...args: HsCTIInitOptions[] | RequiredHsCTIInitOptions[]
    ) {
      // eslint-disable-next-line @typescript-eslint/no-explicit-any
      const currentParams: { [key: string]: any } = args[0]
      let requireList = []
      if (currentParams.scene === Scene.Monitor) {
        requireList = allRequiredParams
      } else {
        requireList = baseRequireParams
      }
      requireList.forEach((param: string) => {
        if (!currentParams[param]) {
          throw `参数[${param}]为必填参数`
        }
      })
      return originalMethod.apply(this, args)
    }
    return descriptor
  }
}
