import { Invitation, Session, UserAgentState } from 'sip.js'
import { LoggerLevels } from '../Logger'
import {
  SocketStatus,
  SIPStatus,
  CTIEvent,
  type BaseOptions,
  type HsCTIInitOptions,
  type RequiredHsCTIInitOptions
} from './outputType'
import {
  SimpleUserDelegate,
  SimpleUserOptions,
  SessionManagerDelegate,
  SessionManagerOptions
} from 'sip.js/lib/platform/web'

/**
 * @enum {string} 本地提示音枚举
 * RingAudio: 来电振铃提示音
 * WaitAudio: 主动外呼等待音
 * ByeAudio: 挂断提示音
 */
export enum AudioName {
  RingAudio = '_ringAudio',
  WaitAudio = '_waitAudio',
  ByeAudio = '_byeAudio'
}

/**
 * @interface {} 初始化 Socket 需要的参数
 * ctiSessionId: CTI 后端生成的单次注册唯一 ID
 * token: 用户 token（一般来说token是必传，一期没有统一的登陆鉴权流程，先不传，后期再完善）
 * imHeartTime: 心跳延迟时间 单位秒
 * imRetryCount: 心跳可以重试的次数
 * loggerLevel: 日志等级
 * imWsServer: IM websocket url
 * imWsPath: IM websocket path
 */
export interface SocketOptions extends BaseOptions {
  ctiSessionId: string | undefined
  token?: string
  imHeartTime: number
  imRetryCount: number
  loggerLevel: LoggerLevels
  imWsServer: string
  imWsPath: string
}

export interface BaseTrackParams {
  agent_id: string
  vcc_id: string
  scene: string
  monitor_scene?: string
}
export interface CtiFlowIdTrackParams {
  cti_flow_id: string
  cti_flow_id_list: string
}

export interface InitConfigTrackParams {
  sip_server: string
  cti_session_id: string
}
export interface MakeCallTrackParams {
  call_id?: string
  called?: string
  circuit_uid?: string
}

export interface SocketBaseTrackParams {
  socket_session_id?: string
}

export interface SocketStatusChangeParams {
  status: SocketStatus
  code?: string
  error?: string
}

export interface SIPStatusChangeParams {
  status: SIPStatus
  code?: string
  error?: string
  method?: string
}

export interface SimpleUserPlusOptions extends SimpleUserOptions {
  delegate?: SimpleUserPlusDelegate
  optionsPingInterval?: number
}

export interface SimpleUserPlusDelegate extends SimpleUserDelegate {
  onCallBegin?(session: Session): void
  onInvite?(invitation: Invitation): void
  onUserAgentStateChange?(state: UserAgentState): void
  onReconnectFailed?(): void
  onReconnectStart?(): void
}

export interface SessionManagerPlusOptions extends SessionManagerOptions {
  delegate?: SessionManagerPlusDelegate
}

export interface SessionManagerPlusDelegate extends SessionManagerDelegate {
  onReconnectFailed?(): void
  onReconnectStart?(): void
}

/**
 * @enum {string} 埋点上报的分类
 * FeSocket: Socket 类暴露的事件
 * FeSIP: SIP 相关事件
 * FeMethod: 业务调用 SDK 暴露的方法
 * FeAPI: SDK 调用后端的接口返回值
 * FeAPIError: SDK 调用后端的接口错误返回值
 * FeIMDown: 后端通过 IM 返回的常规事件
 * FeIMCmd: IM 返回的断开指令
 * FeMedia: SDK 媒体事件
 * FeStatu: SDK 是否状态流转事件
 * FeEmit: SDK 抛出的事件
 */
export enum TrackSource {
  FeSocket = 'fe-socket',
  FeSIP = 'fe-sip',
  FeMethod = 'fe-method',
  FeAPI = 'fe-api',
  FeAPIError = 'fe-api-error',
  FeIMDown = 'fe-im-down',
  FeIMCmd = 'fe-im-cmd',
  FeMedia = 'fe-media',
  FeStatus = 'fe-status',
  FeEmit = 'fe-emit'
}

export interface CTIEventParams {
  eventName: CTIEvent
  ext: { [key: string]: string }
}

export enum SocketEvent {
  SetSocketStatus = 'SetSocketStatus',
  SocketDownEvent = 'SocketDownEvent'
}

export enum ExceptMessage {
  CommonNetworkErrorMsg = '对不起，网络状况暂时不佳，请刷新后重试。',
  CustomNetworkErrorMsg = '对不起，连接失败，请刷新后重试。如果问题持续出现，请联系我们的技术人员。感谢您的理解和支持。',
  ManualCallAnswerErrorMsg = '当前电话未接通，通话已结束，请重试。',
  RobotOrWeChatAnswerErrorMsg = '对不起，由于用户挂机等原因，当前电话未接通，请等待下一通电话。',
  SipByeErrorMsg = '请稍等，正在挂断。',
  CTIRepeatLoginMsg = '当前坐席已被其他终端替代。'
}

export const methodExceptMsgMap: { [key: string]: ExceptMessage } = {
  checkIn: ExceptMessage.CommonNetworkErrorMsg,
  setIdle: ExceptMessage.CustomNetworkErrorMsg,
  setBusy: ExceptMessage.CommonNetworkErrorMsg,
  makeCall: ExceptMessage.CommonNetworkErrorMsg,
  answer: ExceptMessage.CommonNetworkErrorMsg,
  bye: ExceptMessage.CommonNetworkErrorMsg,
  loadAgentGroupData: ExceptMessage.CommonNetworkErrorMsg,
  listen: ExceptMessage.CommonNetworkErrorMsg,
  setActiveService: ExceptMessage.CustomNetworkErrorMsg
}

export const baseRequireParams: (keyof HsCTIInitOptions)[] = [
  'agentId',
  'vccId',
  'password',
  'env',
  'scene'
]

const monitorRequireParams: (keyof RequiredHsCTIInitOptions)[] = [
  'monitorScene'
]

export const allRequiredParams: (keyof RequiredHsCTIInitOptions)[] = [
  ...baseRequireParams,
  ...monitorRequireParams
]
